# -*- coding: utf-8 -*-

__title__ = 'xively-mqtt'
__version__ = '0.1.0-rc2'


__all__ = ['Feed', 'Datastream', 'Datapoint', 'Location', 'Waypoint',
           'Trigger', 'Key', 'Permission', 'Resource']


class Base(object):
    """Abstract base class to store API data and allow (de)serialisation."""

    def __init__(self):
        self._data = {}

    def __getstate__(self):
        """Returns the current state of the object.

        This is the data that should be sent to the Xively API.
        """
        return {k: v for k, v in self._data.items() if v is not None}

    def __getattr__(self, name):
        """Looks up and returns an attribute from the state."""
        try:
            return self._data[name]
        except KeyError:
            class_name = self.__class__.__name__
            raise AttributeError(
                "'{}' object has no attribute '{}'".format(class_name, name))

    def __setattr__(self, name, value):
        """Sets the value of an attribute in the state."""
        if not name.startswith('_') and name not in dir(self.__class__):
            self._data[name] = value
        else:
            super(Base, self).__setattr__(name, value)


class Feed(Base):
    """Xively Feed, which can contain a number of Datastreams.

    :param title: A descriptive name for the feed
    :param description: A longer text description of the feed
    :param website: The URL of a website which is relevant to this feed e.g.
        home page
    :param email: A public contact email address for the provider of this feed
    :param tags: Tagged metadata about the environment (characters ' " and
        commas will be stripped out)
    :param location: :class:`.Location` object for this feed
    :param private: Whether the environment is private or not.
    :type private: bool

    Usage::

        >>> import xively
        >>> xively.Feed(title="Xively Office environment")
        <xively.Feed(None)>

    """

    VERSION = "1.0.0"

    # Set id and feed directly as they aren't part of state. By setting them on
    # the class they won't get entered into _data and will be set on the
    # instance itself.
    id = None
    feed = None

    _datastreams_manager = None

    def __init__(self, title, description=None, website=None, email=None,
                 tags=None, location=None, private=None, datastreams=None):
        """Creates a new Feed."""
        self._data = {
            'version': self.VERSION,
            'title': title,
            'description': description,
            'website': website,
            'email': email,
            'tags': tags,
            'location': location,
            'private': private,
        }
        if datastreams is not None:
            self.datastreams = datastreams

    def __repr__(self):
        return "<{}.{}({id})>".format(
            __package__, self.__class__.__name__, id=self.id)

    @property
    def datastreams(self):
        """Manager for datastreams of this feed.

        When fetched from the API, datastreams behaves like a cache, populated
        with the most recently updated datastreams for this feed. The manager
        can also be used to create, update and delete datastreams for this
        feed.

        Usage::

            >>> import xively
            >>> api = xively.XivelyAPIClient("API_KEY")
            >>> feed = api.feeds.get(7021)
            >>> feed.datastreams[0]  # doctest: +IGNORE_UNICODE
            <xively.Datastream('3')>

        """
        if self._datastreams_manager is None:
            import xively.managers
            self._datastreams_manager = xively.managers.DatastreamsManager(self)
        return self._datastreams_manager

    @datastreams.setter  # NOQA
    def datastreams(self, datastreams):
        manager = getattr(self, '_manager', None)
        if manager:
            # Accessing self.datastreams will create a DatastreamsManager if
            # one didn't already exist.
            manager._coerce_datastreams(datastreams, self.datastreams)
        self._data['datastreams'] = datastreams

    def update(self, fields=None):
        """Updates feed and datastreams via the API.

        If successful, the current datastream values are stored and any changes
        in feed metadata overwrite previous values. Xively stores a server-side
        timestamp in the "updated" attribute and sets the feed to "live" if it
        wasn't before.

        :param fields: If given, only update these fields.
        :type fields: list of strings

        """
        url = self.id
        state = self.__getstate__()
        if fields is not None:
            fields = set(fields)
            state = {k: v for k, v in state.items() if k in fields}
        self._manager.update(url, **state)
   
    def publish(self, fields=None):
        url = self.id
        state = self.__getstate__()
        if fields is not None:
            fields = set(fields)
            state = {k: v for k, v in state.items() if k in fields}
        self._manager.publish(url, **state)

    def subscribe(self, fields=None):
        url = self.id
        state = self.__getstate__()
        self._manager.publish(url, **state)

    def delete(self):
        """Delete this feed via the API.

        .. warning:: This is final and cannot be undone.

        """
        url = self.id
        self._manager.delete(url)


class Datastream(Base):
    """Xively Datastream containing current and historical values.

    :param id: The ID of the datastream
    :param tags: Tagged metadata about the datastream
    :param unit: The :class:`.Unit` of the datastream
    :param min_value: The minimum value since the last reset
    :param max_value: The maximum value since the last reset
    :param current_value: The current value of the datastream
    :param at: The timestamp of the current value
    :param datapoints: A collection of timestamped values

    """

    _datapoints_manager = None

    def __init__(self, id, tags=None, unit=None, min_value=None,
                 max_value=None, current_value=None, datapoints=None, at=None):
        """Creates a new datastream object locally."""
        self._data = {
            'id': id,
            'tags': tags,
            'unit': unit,
            'min_value': min_value,
            'max_value': max_value,
            'current_value': current_value,
            'at': at,
        }
        self.datapoints = datapoints or []

    def __getstate__(self):
        state = super(Datastream, self).__getstate__()
        if not state['datapoints']:
            state.pop('datapoints')
        return state

    def __repr__(self):
        return "<{}.{}({id!r})>".format(
            __package__, self.__class__.__name__, id=self._data.get('id'))

    @property
    def datapoints(self):
        """Manager for datapoints of this datastream.

        When a datastream is fetched with history from the API, datapoints is
        a sequence of timestamped values for the period requested. The manger
        can also be used tocreate, update and delte datapoints for this
        datastream.

        Usage::

            >>> import xively
            >>> import datetime
            >>> api = xively.XivelyAPIClient("API_KEY")
            >>> feed = api.feeds.get(7021)
            >>> datastream = feed.datastreams.get("random5",
            ...     start=datetime.datetime(2013, 1, 1, 14, 0, 0),
            ...     end=datetime.datetime(2013, 1, 1, 16, 0, 0))
            >>> datastream.datapoints[:2]
            ... # doctest: +IGNORE_UNICODE +NORMALIZE_WHITESPACE +ELLIPSIS
            [xively.Datapoint(datetime.datetime(...), '0.25741970'),
             xively.Datapoint(datetime.datetime(...), '0.86826886')]

        """
        if self._datapoints_manager is None:
            import xively.managers
            self._datapoints_manager = xively.managers.DatapointsManager(self)
        return self._datapoints_manager

    @datapoints.setter  # NOQA
    def datapoints(self, datapoints):
        self._data['datapoints'] = datapoints

    def update(self, fields=None):
        """Sends the current state of this datastream to Xively.

        This method updates just the single datastream.

        :param fields: If given, only update these fields.
        :type fields: list of strings

        """
        state = self.__getstate__()
        if fields is not None:
            fields = set(fields)
            state = {k: v for k, v in state.items() if k in fields}
        self._manager.update(self.id, **state)

    def delete(self):
        """Delete this datastream from Xively.

        .. warning:: This is final and cannot be undone.

        """
        self._manager.delete(self.id)


class Datapoint(Base):
    """A Datapoint represents a value at a certain point in time.

    :param at: The timestamp of the datapoint
    :param value: The value at this time

    """

    def __init__(self, at, value):
        """Create a new datapoint locally."""
        super(Datapoint, self).__init__()
        self._data['at'] = at
        self._data['value'] = value

    def __repr__(self):
        classname = 'xively.' + self.__class__.__name__
        return "{}({!r}, {!r})".format(classname, self.at, self.value)

    def update(self):
        """Update this datapoint's value."""
        state = self.__getstate__()
        self._manager.update(state.pop('at'), **state)

    def delete(self):
        """Delete this datapoint.

        .. warning:: This is final and cannot be undone.

        """
        self._manager.delete(self.at)


class Location(Base):
    """The location and location type of a feed.

    :param name: The name of the location
    :param domain: The domain of the location, i.e. 'physical' or 'virtual'
    :param exposure: Whether the location is indoors or outdoors
    :param disposition: Whether the location is mobile or static
    :param lat: The latitude of the feed
    :param lon: The longitude of the feed
    :param ele: The elevation of the feed
    :param waypoints: A list of locations for a mobile feed

    """

    def __init__(self, name=None, domain=None, exposure=None, disposition=None,
                 lat=None, lon=None, ele=None, waypoints=None):
        """Create a local location instance."""
        self._data = {
            'name': name,
            'domain': domain,
            'exposure': exposure,
            'disposition': disposition,
            'lat': lat,
            'lon': lon,
            'ele': ele,
        }
        if waypoints is not None:
            self._data['waypoints'] = waypoints


class Waypoint(Base):
    """A waypoint represents where a mobile feed was at a particular time.

    :param at: The timestamp of the waypoint
    :param lat: The latitude at that time
    :param lon: The longitude at that time

    """

    def __init__(self, at, lat, lon):
        """Create a location waypoint, a timestamped cordinate pair."""
        self._data = {
            'at': at,
            'lat': lat,
            'lon': lon,
        }


class Unit(Base):
    """A type, label and symbol of a values unit."""

    def __init__(self, label=None, type=None, symbol=None):
        self._data = {
            'label': label,
            'type': type,
            'symbol': symbol,
        }


class Trigger(Base):
    """Triggers provide 'push' capabilities (aka notifications).

    To create a new trigger, use the :class:`~xively.api.TriggersManager` on a
    :class:`~xively.api.XivelyAPIClient` instance.

    >>> from xively import XivelyAPIClient
    >>> api = XivelyAPIClient("API_KEY")
    >>> api.triggers.create(123, "temperature", "http://example.com", "frozen")
    <xively.Trigger(3)>

    """

    def __init__(self, environment_id, stream_id, url, trigger_type,
                 threshold_value=None):
        self._data = {
            'environment_id': environment_id,
            'stream_id': stream_id,
            'url': url,
            'trigger_type': trigger_type,
        }
        if threshold_value is not None:
            self._data['threshold_value'] = threshold_value

    def __repr__(self):
        return "<{}.{}({id!r})>".format(
            __package__, self.__class__.__name__, id=self.id)

    def update(self, fields=None):
        """Update an existing trigger.

        :param fields: If given, only update these fields
        :type fields: list of strings

        """
        state = self.__getstate__()
        state.pop('id', None)
        if fields is not None:
            fields = set(fields)
            state = {k: v for k, v in state.items() if k in fields}
        self._manager.update(self.id, **state)

    def delete(self):
        """Delete a trigger.

        .. warning:: This is final and cannot be undone.

        """
        self._manager.delete(self.id)


class Key(Base):
    """Keys set which permissions are granted for certain resources.

    :param label: A label by which the key can be referenced
    :param permissions: Collection of Permission objects controlling the access
                        level
    :param expires_at: Expiry date for the key after which it won't work
    :param private_access: Flag that indicates whether this key can access
                           private resources belonging to the user

    """

    def __init__(self, label, permissions, expires_at=None,
                 private_access=False):
        self._data = {
            'label': label,
            'permissions': permissions,
            'private_access': private_access,
        }
        if expires_at:
            self._data['expires_at'] = expires_at

    def __repr__(self):
        return "<{}.{}({label!r})>".format(
            __package__, self.__class__.__name__, label=self.label)

    def delete(self):
        """Delete this key."""
        self._manager.delete(self.api_key)


class Permission(Base):
    """Permissions restrict what can be done by a key.

    :param access_methods:
        A list containing one or more of [get, put, post, delete] indicating
        what type of access the key has
    :param source_ip:
        An IP address that access should be restricted to, so if specified,
        only requests coming from this IP address will be permitted
    :param referer:
        The referer domain. If present this key will only be able to be
        embedded in a web page with the matching URL.  Subdomains are treated
        as different domains
    :param minimum_interval:
        Can be used to create a key that can only request data with a certain
        resolution, i.e. a key could be created that only displays graphs with
        daily values when embedded in a web page. The same key could not be
        used to access full resolution data
    :param label:
        Optional label for identifying permission set
    :param resources:
        Optional collection of Resource objects restricting access to specific
        feeds or datastreams

    """

    def __init__(self, access_methods, source_ip=None, referer=None,
                 minimum_interval=None, label=None, resources=None):
        self._data = {
            'access_methods': access_methods,
        }
        self._data.update((name, value) for (name, value) in (
            ('source_ip', source_ip),
            ('referer', referer),
            ('minimum_interval', minimum_interval),
            ('label', label),
            ('resources', resources),
        ) if value is not None)


class Resource(Base):
    """A Resource defines what an API key has access to.

    :param feed_id:
        Reference to a specific feed id
    :param datastream_id:
        Reference to a specific datastream id within a feed. If specified then
        the feed id must also be specified

    """

    def __init__(self, feed_id, datastream_id=None):
        self._data = {
            'feed_id': feed_id,
        }
        if datastream_id:
            self._data['datastream_id'] = datastream_id



# -*- coding: utf-8 -*-

__title__ = 'xively-mqtt'
__version__ = '0.1.0-rc2'

__all__ = ['Client', 'XivelyAPIClient', 'Datapoint', 'Datastream', 'Feed', 'Key',
           'Location', 'Permission', 'Resource', 'Trigger', 'Unit', 'Waypoint']

from xively-mqtt.api import XivelyAPIClient
from xively-mqtt.client import Client
from xively-mqtt.models import (
    Datapoint, Datastream, Feed, Key, Location, Permission, Resource, Trigger,
    Unit, Waypoint)


def setup_module(module):
    import mock
    import fixtures
    patcher = mock.patch('xively-mqtt.client.Session.request')
    mock_request = patcher.start()
    mock_request.side_effect = fixtures.handle_request
    module._patcher = patcher


def teardown_module(module):
    module._patcher.stop()



# -*- coding: utf-8 -*-

try:
    from urlparse import urljoin
except ImportError:
    from urllib.parse import urljoin  # NOQA

from xively-mqtt.client import Client
from xively-mqtt.managers import FeedsManager, KeysManager, TriggersManager


__all__ = ['XivelyAPIClient']


class XivelyAPIClient(object):
    """An instance of an authenticated Xively API Client.

    The root object from which the user can manage feeds, keys and triggers.

    :param key: A Xively API Key
    :type key: str
    :param use_ssl: Use https for all connections instead of http
    :type use_ssl: bool [False]
    :param kwargs: Other additional keyword arguments to pass to client

    Usage::

        >>> import xively
        >>> xively.XivelyAPIClient("API_KEY")
        <xively.XivelyAPIClient()>

        >>> api = xively.XivelyAPIClient("API_KEY", use_ssl=True)
        >>> api.feeds.base_url
        'https://api.xively.com/v2/feeds'
        >>> api.triggers.base_url
        'https://api.xively.com/v2/triggers'
        >>> api.keys.base_url
        'https://api.xively.com/v2/keys'

    """
    api_version = 'v2'
    client_class = Client

    def __init__(self, key, use_ssl=False, **kwargs):
        self.client = self.client_class(key, use_ssl=use_ssl, **kwargs)
        self.client.base_url += '/{}/'.format(self.api_version)
        self._feeds = FeedsManager(self.client)
        self._triggers = TriggersManager(self.client)
        self._keys = KeysManager(self.client)

    def __repr__(self):
        return "<{}.{}()>".format(__package__, self.__class__.__name__)

    @property
    def feeds(self):
        """
        Access :class:`.Feed` objects through a :class:`.FeedsManager`.
        """
        return self._feeds

    @property
    def triggers(self):
        """
        Access :class:`.Trigger` objects through a :class:`.TriggersManager`.
        """
        return self._triggers

    @property
    def keys(self):
        """
        Access :class:`.Key` objects through a :class:`.KeysManager`.
        """
        return self._keys



from setuptools import setup, find_packages
import sys, os

version = '0.1'

setup(name='xively-mqtt',
      version=version,
      description="A port of Xively library that uses MQTT instead of HTTP as data transport",
      long_description="""\
This project tries to implement a client library based on Python for Xively IoT platform. Using MQTT is expected to increase response time and save costly communication bandwidth""",
      classifiers=[], # Get strings from http://pypi.python.org/pypi?%3Aaction=list_classifiers
      keywords='mqtt xively iot m2m',
      author='Bontor Humala',
      author_email='bontorhumala@gmail.com',
      url='https://github.com/bontorhumala/xively-python',
      license='GPL',
      packages=['xively-mqtt'],
      include_package_data=True,
      zip_safe=True,
      install_requires=[
          # -*- Extra requirements: -*-
          'requests >= 1.1.0',
          'paho-mqtt >= 0.9.1',
      ],
      entry_points="""
      # -*- Entry points: -*-
      """,
      )



# -*- coding: utf-8 -*-

from collections import Sequence
from datetime import datetime

try:
    from urlparse import urljoin
except ImportError:
    from urllib.parse import urljoin  # NOQA

from xively-mqtt.models import (
    Datapoint,
    Datastream,
    Feed,
    Key,
    Location,
    Permission,
    Resource,
    Trigger,
    Unit,
    Waypoint,
)


class ManagerBase(object):
    """Abstract base class for all of out manager classes."""

    @property
    def base_url(self):
        if getattr(self, '_base_url', None) is not None:
            return self._base_url
        parent = getattr(self, 'parent', None)
        if parent is None:
            return
        manager = getattr(parent, '_manager', None)
        if manager is None:
            return
        base_url = manager.url(parent.id) + '/' + self.resource
        return base_url

    @base_url.setter  # NOQA
    def base_url(self, base_url):
        self._base_url = base_url

    def url(self, id_or_url=None):
        """Return a url relative to the base url."""
        url = self.base_url
        if id_or_url:
            url = urljoin(url + '/', str(id_or_url))
        return url

    def _parse_datetime(self, value):
        """Parse and return a datetime string from the Xively API."""
        return datetime.strptime(value, "%Y-%m-%dT%H:%M:%S.%fZ")

    def _prepare_params(self, params):
        """Prepare parameters to be passed in query strings to the Xively API."""
        params = dict(params)
        for name, value in params.items():
            if isinstance(value, datetime):
                params[name] = value.isoformat() + 'Z'
        return params


class FeedsManager(ManagerBase):
    """Create, update and return Feed objects.

    .. note:: This manager should live on a :class:`.XivelyAPIClient` instance
        and not instantiated directly.

    :param client: Low level :class:`.Client` instance

    Usage::

        >>> import xively
        >>> api = xively.XivelyAPIClient("API_KEY")
        >>> api.feeds.create(title="Xively Office environment")
        <xively.Feed(7021)>
        >>> api.feeds.get(7021)
        <xively.Feed(7021)>
        >>> api.feeds.update(7021, private=True)
        >>> api.feeds.delete(7021)

    """

    resource = 'feeds'
    mqttc = self.client.mqttc
    mqttc.on_publish = on_publish

    # List of fields that can be returned from the API but not directly set.
    _readonly_fields = (
        'id',
        'feed',
        'status',
        'creator',
        'created',
        'updated',
        'version',
        'auto_feed_url',
        'product_id',
        'device_serial',
    )

    def __init__(self, client):
        self.client = client
        self.base_url = client.base_url + self.resource

    def create(self, title, description=None, website=None, email=None,
               tags=None, location=None, private=None, datastreams=None):
        """Creates a new Feed.

        :param title: A descriptive name for the feed
        :param description: A longer text description of the feed
        :param website: The URL of a website which is relevant to this feed
            e.g. home page
        :param email: A public contact email address for the creator of this
            feed
        :param tags: Tagged metadata about the environment (characters ' " and
            commas will be stripped out)
        :param location: :class:`.Location` object for this feed
        :param private: Whether the environment is private or not. Can be
            either True or False

        """
        data = {
            'version': Feed.VERSION,
            'title': title,
            'description': description,
            'website': website,
            'email': email,
            'tags': tags,
            'location': location,
            'private': private,
            'datastreams': datastreams,
        }
        feed = self._coerce_feed(data)
        response = self.client.post(self.url(), data=feed)
        response.raise_for_status()
        location = response.headers['location']
        feed.feed = location
        feed.id = _id_from_url(location)
        return feed

    def update(self, id_or_url, **kwargs):
        """Updates an existing feed by its id or url.

        :param id_or_url: The id of a :class:`.Feed` or its URL
        :param kwargs: The fields to be updated

        """
        url = self.url(id_or_url)
        response = self.client.put(url, data=kwargs)
        response.raise_for_status()
 
    def on_publish(client, userdata, mid):
        print("Message " + str(mid) + " has been published.")

    def publish(self, id_or_url, **kwargs):
        """Publish message on a feed by its id or url.

        :param id_or_url: The id of a :class:`.Feed` or its URL
        :param kwargs: The fields to be updated

        """
        topic = '/v2/feeds/' + id_or_url + '.json'
        mqttc.publish(topic=topic, payload=kwargs, qos=mqtt.qos, retain=mqtt.retain)

    def subscribe(self, id_or_url):
        topic = '/v2/feeds/' + id_or_url + '.json'
        mqttc.subscribe(topic=topic, qos=mqtt.qos)

    def list(self, page=None, per_page=None, content=None, q=None, tag=None,
             user=None, units=None, status=None, order=None, show_user=None,
             lat=None, lon=None, distance=None, distance_units=None):
        """Returns a paged list of feeds.

        Only feeds that are viewable by the authenticated account will be
        returned. The following parameters can be applied to limit or refine
        the returned feeds:

        :param page: Integer indicating which page of results you are
            requesting. Starts from 1.
        :param per_page: Integer defining how many results to return per page
            (1 to 1000)
        :param content: String parameter ('full' or 'summary') describing
            whether we want full or summary results. Full results means all
            datastream values are returned, summary just returns the
            environment meta data for each feed
        :param q: Full text search parameter. Should return any feeds matching
            this string
        :param tag: Returns feeds containing datastreams tagged with the search
            query
        :param user: Returns feeds created by the user specified
        :param units: Returns feeds containing datastreams with units specified
            by the search query
        :param status: Possible values ('live', 'frozen', or 'all'). Whether to
            search for only live feeds, only frozen feeds, or all feeds.
            Defaults to all
        :param order: Order of returned feeds. Possible values ('created_at',
            'retrieved_at', or 'relevance')
        :param show_user: Include user login and user level for each feed.
            Possible values: true, false (default)

        The following additional parameters are available which allow location
        based searching of feeds:

        :param lat: Used to find feeds located around this latitude
        :param lon: Used to find feeds located around this longitude
        :param distance: search radius
        :param distance_units: miles or kms (default)

        """
        url = self.url()
        params = {k: v for k, v in (
            ('page', page),
            ('per_page', per_page),
            ('content', content),
            ('q', q),
            ('tag', tag),
            ('user', user),
            ('units', units),
            ('status', status),
            ('order', order),
            ('show_user', show_user),
            ('lat', lat),
            ('lon', lon),
            ('distance', distance),
            ('distance_units', distance_units),
        ) if v is not None}
        response = self.client.get(url, params=params)
        response.raise_for_status()
        json = response.json()
        feeds = [self._coerce_feed(feed_data) for feed_data in json['results']]
        return feeds

    def get(self, id_or_url, datastreams=None, show_user=None, start=None,
            end=None, duration=None, find_previous=None, limit=None,
            interval_type=None, interval=None):
        """Fetches and returns a feed by id or url.

        By default the most recent datastreams are returned. It is also
        possible to filter the datastreams returned with the feed by using the
        "datastreams" parameter and a list of datastream IDs.

        :param datastreams: Filter the returned datastreams
        :type datastreams: list of datastream IDs
        :param show_user: Include user login for each feed. (default: False)
        :type show_user: bool

        :param start: Defines the starting point of the query
        :param end: Defines the end point of the data returned
        :param duration: Specifies the duration of the query
        :param find_previous:
            Will also return the previous value to the date range being
            requested.
        :param limit:
            Limits the number of results to the number specified.  Defaults to
            100 and has a maximum of 1000.
        :param interval_type:
            If set to "discrete" the data will be returned in fixed time
            interval format according to the inverval value supplied. If this
            is not set, the raw datapoints will be returned.
        :param interval:
            Determines what interval of data is requested and is defined in
            seconds between the datapoints. If a value is passed in which does
            not match one of these values, it is rounded up to the next value.

        See :meth:`~.DatapointsManager.history` for details.

        """
        url = self.url(id_or_url)
        if isinstance(datastreams, Sequence):
            datastreams = ','.join(datastreams)
        params = {k: v for k, v in (
            ('datastreams', datastreams),
            ('show_user', show_user),
            ('start', start),
            ('end', end),
            ('duration', duration),
            ('find_previous', find_previous),
            ('limit', limit),
            ('interval_type', interval_type),
            ('interval', interval),
        ) if v is not None}
        params = self._prepare_params(params)
        response = self.client.get(url, params=params)
        response.raise_for_status()
        data = response.json()
        feed = self._coerce_feed(data)
        return feed

    def delete(self, id_or_url):
        """Delete a feed by id or url.

        .. WARNING:: This is final and cannot be undone.

        :param id_or_url: The feed ID  or its URL

        """
        url = self.url(id_or_url)
        response = self.client.delete(url)
        response.raise_for_status()

    def _coerce_feed(self, feed_data):
        """Returns a Feed object from a mapping object (dict)."""
        datastreams_data = feed_data.pop('datastreams', None)
        location_data = feed_data.pop('location', None)
        # Strip out the readonly fields and manually set later.
        readonly = {f: feed_data.pop(f)
                    for f in self._readonly_fields
                    if f in feed_data}
        feed = Feed(**feed_data)
        feed._manager = self
        feed.id = readonly.pop('id', None)
        feed.feed = readonly.pop('feed', None) or self.url(feed.id)
        # Explicitely set the readonly fields we stripped out earlier.
        for name, value in readonly.items():
            setattr(feed, name, value)
        if datastreams_data:
            feed._datastreams_manager = DatastreamsManager(feed)
            feed.datastreams = self._coerce_datastreams(
                datastreams_data, feed._datastreams_manager)
        if location_data:
            location = self._coerce_location(location_data)
        else:
            location = Location()
        feed._data['location'] = location
        return feed

    def _coerce_datastreams(self, datastreams_data, datastreams_manager):
        """Returns Datastream objects from the data given."""
        datastreams = []
        for data in datastreams_data:
            datastream = datastreams_manager._coerce_datastream(data)
            datastreams.append(datastream)
        return datastreams

    def _coerce_location(self, instance):
        """Returns a Location object, converted from instance if required."""
        if isinstance(instance, Location):
            location = instance
        else:
            location_data = dict(**instance)
            waypoints_data = location_data.pop('waypoints', None)
            if waypoints_data is not None:
                waypoints = self._coerce_waypoints(waypoints_data)
                location_data['waypoints'] = waypoints
            location = Location(**location_data)
        return location

    def _coerce_waypoints(self, waypoints_data):
        """Returns a list of Waypoint objects from the given waypoint data."""
        waypoints = []
        for data in waypoints_data:
            at = self._parse_datetime(data['at'])
            data = {k: v for k, v in data.items() if k != 'at'}
            waypoint = Waypoint(at=at, **data)
            waypoints.append(waypoint)
        return waypoints


class DatastreamsManager(Sequence, ManagerBase):
    """Create, update and return Datastream objects.

    Instances of this class hang off of :class:`.Feed` objects to manage
    datastreams of that feed.

    A list of datastreams can be retrieved along with the feed which can be
    accessed via this instance as a sequence.

    :param feed: A :class:`.Feed` instance.

    Usage::

        >>> import xively
        >>> api = xively.XivelyAPIClient("API_KEY")
        >>> feed = api.feeds.get(7021)
        >>> list(feed.datastreams)  # doctest: +IGNORE_UNICODE
        [<xively.Datastream('3')>, <xively.Datastream('4')>]
        >>> feed.datastreams.create("1")
        <xively.Datastream('1')>

    """

    resource = 'datastreams'

    # List of fields that can be returned from the API but not directly set.
    _readonly_fields = (
        'at',
        'current_value',
    )

    def __init__(self, feed):
        self.parent = feed
        feed_manager = getattr(feed, '_manager', None)
        self.client = getattr(feed_manager, 'client', None)

    def __contains__(self, value):
        return value in self.datastreams['datastreams']

    def __getitem__(self, item):
        return self._datastreams[item]

    def __len__(self):
        return len(self._datastreams)

    @property
    def _datastreams(self):
        return self.parent._data.setdefault('datastreams', [])

    def create(self, id, current_value=None, tags=None, unit=None,
               min_value=None, max_value=None, at=None):
        """Creates a new datastream on a feed.

        :param id: The ID of the datastream
        :param current_value: The current value of the datastream
        :param tags: Tagged metadata about the datastream
        :param unit: The :class:`.Unit` for this datastream
        :param min_value: The minimum value since the last reset
        :param max_value: The maximum value since the last reset
        :param at: The timestamp of the current value
        :returns: A :class:`.Datastream` object

        """
        datastream_data = dict(
            id=id,
            current_value=current_value,
            tags=tags,
            unit=unit,
            min_value=min_value,
            max_value=max_value,
            at=at)
        datastream = self._coerce_datastream(datastream_data)
        data = {
            'version': self.parent.version,
            'datastreams': [datastream],
        }
        response = self.client.post(self.url(), data=data)
        response.raise_for_status()
        return datastream

    def update(self, datastream_id, **kwargs):
        """Updates a feeds datastream by id.

        :param datastream_id: The ID of the datastream to update
        :param kwargs: The datastream fields to be updated

        """
        url = self.url(datastream_id)
        response = self.client.put(url, data=kwargs)
        response.raise_for_status()

    def list(self, datastreams=None, show_user=None):
        """Returns a list of datastreams for the parent feed object.

        :param datastreams: Filter the returned datastreams
        :type datastreams: list of datastream IDs
        :param show_user: Include user login for each feed (default: False)
        :type show_user: bool

        """
        url = self.url('..')
        params = {k: v for k, v in (
            ('datastreams', datastreams),
            ('show_user', show_user),
        ) if v is not None}
        response = self.client.get(url, params=params)
        response.raise_for_status()
        json = response.json()
        for datastream_data in json.get('datastreams', []):
            datastream = self._coerce_datastream(datastream_data)
            yield datastream

    def get(self, id_or_url, start=None, end=None, duration=None,
            find_previous=None, limit=None, interval_type=None, interval=None):
        """Fetches and returns a feed's datastream by its id.

        If start, end or duration are given, also returns Datapoints for that
        period.

        :param id_or_url: The ID of the datastream to retrieve or its URL
        :param start: Defines the starting point of the query
        :param end: Defines the end point of the data returned
        :param duration: Specifies the duration of the query
        :param find_previous:
            Will also return the previous value to the date range being
            requested.
        :param limit:
            Limits the number of results to the number specified.  Defaults to
            100 and has a maximum of 1000.
        :param interval_type:
            If set to "discrete" the data will be returned in fixed time
            interval format according to the inverval value supplied. If this
            is not set, the raw datapoints will be returned.
        :param interval:
            Determines what interval of data is requested and is defined in
            seconds between the datapoints. If a value is passed in which does
            not match one of these values, it is rounded up to the next value.

        See :meth:`~.DatapointsManager.history` for details.

        """
        url = self.url(id_or_url)
        params = {k: v for k, v in (
            ('start', start),
            ('end', end),
            ('duration', duration),
            ('find_previous', find_previous),
            ('limit', limit),
            ('interval_type', interval_type),
            ('interval', interval),
        ) if v is not None}
        params = self._prepare_params(params)
        response = self.client.get(url, params=params)
        response.raise_for_status()
        data = response.json()
        datastream = self._coerce_datastream(data)
        return datastream

    def delete(self, id_or_url):
        """Delete a datastream by id or url.

        .. WARNING:: This is final and cannot be undone.

        :param id_or_url: The datastream ID or its URL

        """
        url = self.url(id_or_url)
        response = self.client.delete(url)
        response.raise_for_status()

    def _coerce_datapoints(self, datapoints_manager, datapoints_data):
        """Returns Datapoints objects from a list of mapping objects (dict)."""
        datapoints = []
        for data in datapoints_data:
            data['at'] = self._parse_datetime(data['at'])
            datapoint = datapoints_manager._coerce_datapoint(data)
            datapoints.append(datapoint)
        return datapoints

    def _coerce_unit(self, instance):
        """Returns a Unit object, converted from instance if required."""
        if isinstance(instance, Unit):
            unit = instance
        else:
            instance_data = dict(**instance)
            unit = Unit(**instance_data)
        return unit

    def _coerce_datastream(self, d):
        """Returns a Datastream object from a mapping object (dict)."""
        if isinstance(d, dict):
            datapoints_data = d.pop('datapoints', None)
            unit_data = d.pop('unit', None)
            # Remove version, part of Feed not Datastream
            d.pop('version', None)
            # Strip out the readonly fields and manually set later.
            readonly = {f: d.pop(f) for f in self._readonly_fields if f in d}
            datastream = Datastream(**d)
            # Explicitely set the readonly fields we stripped out earlier.
            for name, value in readonly.items():
                setattr(datastream, name, value)
            if datapoints_data:
                datapoints = self._coerce_datapoints(
                    datastream.datapoints, datapoints_data)
                datastream.datapoints = datapoints
            if unit_data:
                unit = self._coerce_unit(unit_data)
                datastream.unit = unit
        elif isinstance(d, Datastream):
            datastream = d
        datastream._manager = self
        return datastream


class DatapointsManager(Sequence, ManagerBase):
    """Manage datapoints of a datastream.

    A list of :class:`.Datapoint` objects can be retrieved along with the
    :class:`.Datastream` (or :class:`.Feed`) which can be accessed via this
    instance as a sequence.

    :param datastream: A :class:`.Datastream` instance.

    """

    resource = 'datapoints'

    def __init__(self, datastream):
        self.parent = datastream
        datastream_manager = getattr(datastream, '_manager', None)
        self.client = getattr(datastream_manager, 'client', None)

    def __contains__(self, value):
        return value in self.datapoints['datapoints']

    def __getitem__(self, item):
        return self._datapoints[item]

    def __len__(self):
        return len(self._datapoints)

    @property
    def _datapoints(self):
        return self.parent._data['datapoints']

    def create(self, value, at=None):
        """Create a single new datapoint for this datastream.

        :param at: The timestamp of the datapoint (default: datetime.now())
        :param value: The value at this time

        To create multiple datapoints at the same time do the following
        instead:

        .. note:: You can use ISO8601 formatted strings instead of datetime
                  objects when dealing with the API.

        >>> import xively
        >>> api = xively.XivelyAPIClient("API_KEY")
        >>> feed = api.feeds.get(7021)
        >>> datastream = feed.datastreams[0]
        >>> # First create the datapoints.
        >>> datastream.datapoints = [
        ...     xively.Datapoint(at="2010-05-20T11:01:43Z", value=294),
        ...     xively.Datapoint(at="2010-05-20T11:01:44Z", value=295),
        ...     xively.Datapoint(at="2010-05-20T11:01:45Z", value=296),
        ...     xively.Datapoint(at="2010-05-20T11:01:46Z", value=297),
        ... ]
        >>> # Then send them to the server.
        >>> datastream.update(fields='datapoints')

        """
        at = at or datetime.now()
        datapoint = Datapoint(at, value)
        payload = {'datapoints': [datapoint]}
        response = self.client.post(self.url(), data=payload)
        response.raise_for_status()
        return datapoint

    def update(self, at, value):
        """Update the value of a datapiont at a given timestamp.

        :param at: The timestamp of a value to change
        :param value: The value to change

        .. note:: A datapoint at the given time must already exist.

        """
        url = "{}/{}Z".format(self.url(), at.isoformat())
        payload = {'value': value}
        response = self.client.put(url, data=payload)
        response.raise_for_status()

    def get(self, at):
        """Fetch and return a :class:`.Datapoint` at the given timestamp.

        :param at: The timestamp to return a datapoint for

        """
        url = "{}/{}Z".format(self.url(), at.isoformat())
        response = self.client.get(url)
        response.raise_for_status()
        data = response.json()
        data['at'] = self._parse_datetime(data['at'])
        return self._coerce_datapoint(data)

    def history(self, start=None, end=None, duration=None, find_previous=None,
                limit=None, interval_type=None, interval=None):
        """Fetch and return a list of datapoints in a given timerange.

        :param start: Defines the starting point of the query
        :param end: Defines the end point of the data returned
        :param duration: Specifies the duration of the query
        :param find_previous:
            Will also return the previous value to the date range being
            requested.
        :param limit:
            Limits the number of results to the number specified.  Defaults to
            100 and has a maximum of 1000.
        :param interval_type:
            If set to "discrete" the data will be returned in fixed time
            interval format according to the inverval value supplied. If this
            is not set, the raw datapoints will be returned.
        :param interval:
            Determines what interval of data is requested and is defined in
            seconds between the datapoints. If a value is passed in which does
            not match one of these values, it is rounded up to the next value.

        .. note::

            ``find_previous`` is useful for any graphing because if you
            want to draw a graph of the date range you specified you would end
            up with a small gap until the first value.

        .. note::

            In order to paginate through the data use the last timestamp
            returned as the start of the next query.

        .. note::

            The maximum number of datapoints able to be returned from the API
            in one query is 1000. If you need more than 1000 datapoints for
            a specific period you should use the start and end times to split
            them up into smaller chunks.

        The valid time units are::

        * seconds
        * minute(s)
        * hour(s)
        * day(s)
        * week(s)
        * month(s)
        * year(s)

        The acceptable intervals are currently:

        ===== ============================== ==========================
        Value Description                    Maximum range in one query
        ===== ============================== ==========================
        0     Every datapoint stored         6 hours
        30    One datapoint every 30 seconds 12 hours
        60    One datapoint every minute     24 hours
        300   One datapoint every 5 minutes  5 days
        900   One datapoint every 15 minutes 14 days
        1800  One datapoint per 30 minutes   31 days
        3600  One datapoint per hour         31 days
        10800 One datapoint per three hours  90 days
        21600 One datapoint per six hours    180 days
        43200 One datapoint per twelve hours 1 year
        86400 One datapoint per day          1 year
        ===== ============================== ==========================

        """
        url = self.url('..').rstrip('/')
        params = {k: v for k, v in (
            ('start', start),
            ('end', end),
            ('duration', duration),
            ('find_previous', find_previous),
            ('limit', limit),
            ('interval_type', interval_type),
            ('interval', interval),
        ) if v is not None}
        params = self._prepare_params(params)
        response = self.client.get(url, params=params)
        response.raise_for_status()
        data = response.json()
        for datapoint_data in data.get('datapoints', []):
            datapoint_data['at'] = self._parse_datetime(datapoint_data['at'])
            yield self._coerce_datapoint(datapoint_data)

    def delete(self, at=None, start=None, end=None, duration=None):
        """Delete a datapoint or a range of datapoints.

        :param at: A timestamp of a single datapoint to delete
        :param start: Defines the starting point of the query
        :param end: Defines the end point of the datapoints deleted
        :param duration: Specifies the duration of the query

        By providing a start and end timestamp as query parameters, you may
        remove all datapoints that lie between those dates. If you send your
        request with only a start timestamp, all datapoints after the value
        will be removed. Providing an end timestamp will remove all datapoints
        prior to the supplied value.

        Additionally, this method supports a duration parameter (e.g.
        ``duration="3hours"``). Providing a `start` and a `duration` will
        delete all datapoints between the `start` and (`start` + `duration`).
        Providing `end` will delete all datapoints between (`end` - `duration`)
        and `end`. The formatting of the `duration` parameter is the same as is
        used in the :meth:`.history` method.

        .. warning: This is final and cannot be undone.

        """
        url = self.url()
        params = {k: v for k, v in (
            ('start', start),
            ('end', end),
            ('duration', duration),
        ) if v is not None}
        if at:
            url = "{}/{}Z".format(url, at.isoformat())
        elif params:
            params = self._prepare_params(params)
        response = self.client.delete(url, params=params)
        response.raise_for_status()

    def _coerce_datapoint(self, d):
        if isinstance(d, Datapoint):
            datapoint = self._clone_datapoint(d)
        elif isinstance(d, dict):
            datapoint = Datapoint(**d)
        datapoint._manager = self
        return datapoint

    def _clone_datapoint(self, d):
        return Datapoint(**d._data)


class TriggersManager(ManagerBase):
    """Manage :class:`.Trigger`.

    This manager should live on a :class:`.XivelyAPIClient` instance and not
    instantiated directly.

    :param client: Low level :class:`.Client` instance

    Usage::

        >>> import xively
        >>> api = xively.XivelyAPIClient("API_KEY")
        >>> api.triggers.create(
        ...     environment_id=8470, stream_id="0",
        ...     url="http://www.postbin.org/1ijyltn",
        ...     trigger_type='lt', threshold_value="15.0")
        <xively.Trigger(3)>

    """

    resource = 'triggers'

    _readonly_fields = (
        'id',
        'notified_at',
        'user',
    )

    def __init__(self, client):
        self.client = client
        self.base_url = client.base_url + self.resource

    def create(self, environment_id, stream_id, url, trigger_type,
               threshold_value=None):
        """Create a new :class:`.Trigger`.

        :param environment_id: An ID of a :class:`.Feed`
        :param stream_id: An ID of a :class:`.Datastream`
        :param url: The URL to POST events to
        :param trigger_type: The type of trigger (from below)
        :param threshold_value: The threshold at which the trigger fires

        :returns: A new :class:`.Trigger` object.

        Possible values for ``trigger_type`` are:

        ======= ================================
        gt      greater than
        gte     greater than or equal to
        lt      less than
        lte     less than or equal to
        eq      equal to
        change  any change
        frozen  no updates for 15 minutes
        live    updated again after being frozen
        ======= ================================

        """
        data = {
            'environment_id': environment_id,
            'stream_id': stream_id,
            'url': url,
            'trigger_type': trigger_type,
            'threshold_value': threshold_value,
        }
        trigger = self._coerce_trigger(data)
        response = self.client.post(self.url(), data=trigger)
        response.raise_for_status()
        trigger._manager = self
        location = response.headers['location']
        trigger._data['id'] = int(location.rsplit('/', 1)[1])
        return trigger

    def get(self, id_or_url):
        """Fetch and return an existing trigger.

        :param id_or_url: The ID of the trigger or its URL

        """
        url = self.url(id_or_url)
        response = self.client.get(url)
        response.raise_for_status()
        data = response.json()
        data.pop('id')
        notified_at = data.pop('notified_at', None)
        user = data.pop('user', None)
        trigger = self._coerce_trigger(data)
        trigger._data['id'] = id_or_url
        if notified_at:
            trigger._data['notified_at'] = self._parse_datetime(notified_at)
        if user:
            trigger._data['user'] = user
        trigger._manager = self
        return trigger

    def update(self, id_or_url, **kwargs):
        """Update an existing trigger.

        :param id_or_url: The ID of the :class:`.Trigger` to update or its URL
        :param kwargs: The fields to be updated

        """
        url = self.url(id_or_url)
        response = self.client.put(url, data=kwargs)
        response.raise_for_status()

    def list(self, feed_id=None):
        """Return a list of triggers.

        :param feed_id: Filter the returned triggers to only include those on
                        datastreams of the specified feed.

        """
        url = self.url()
        params = {k: v for k, v in (
            ('feed_id', feed_id),
        ) if v is not None}
        response = self.client.get(url, params=params)
        response.raise_for_status()
        json = response.json()
        for data in json:
            trigger = self._coerce_trigger(data)
            trigger._manager = self
            yield trigger

    def delete(self, id_or_url):
        """Delete a trigger by id or url.

        .. WARNING:: This is final and cannot be undone.

        :param id_or_url: The datastream ID or its URL

        """
        url = self.url(id_or_url)
        response = self.client.delete(url)
        response.raise_for_status()

    def _coerce_trigger(self, d):
        # Strip out the readonly fields and manually set later.
        readonly = {f: d.pop(f) for f in self._readonly_fields if f in d}
        trigger = Trigger(**d)
        # Explicitely set the readonly fields we stripped out earlier.
        for name, value in readonly.items():
            setattr(trigger, name, value)
        return trigger


class KeysManager(ManagerBase):
    """Manage keys their permissions and restrict by resource.

    This manager should live on a :class:`.XivelyAPIClient` instance and not
    instantiated directly.

    :param client: Low level :class:`.Client` instance

    Usage::

        >>> import xively
        >>> api = xively.XivelyAPIClient("API_KEY")
        >>> api.keys.create(
        ...     label="sharing key",
        ...     private_access=True,
        ...     permissions=[
        ...         xively.Permission(
        ...             access_methods=["put"],
        ...             source_ip="128.44.98.129",
        ...             resources=[
        ...                 xively.Resource(feed_id=504),
        ...             ]),
        ...         xively.Permission(access_methods=["get"])
        ...     ])
        <xively.Key('sharing key')>

    """

    resource = 'keys'

    def __init__(self, client):
        self.client = client
        self.base_url = client.base_url + self.resource

    def create(self, label, permissions, expires_at=None, private_access=None):
        """Create a new API key.

        :param label: A label by which the key can be referenced
        :param permissions: Collection of Permission objects controlling the
                            access level
        :param expires_at: Expiry date for the key after which it won't work
        :param private_access: Flag that indicates whether this key can access
                               private resources belonging to the user

        """
        data = dict(label=label, permissions=permissions,
                    expires_at=expires_at, private_access=private_access)
        key = self._coerce_key(data)
        response = self.client.post(self.url(), data={'key': key})
        response.raise_for_status()
        location = response.headers['Location']
        key.api_key = _id_from_url(location)
        return key

    def list(self, feed_id=None):
        """List all API keys for this account or for the given feed.

        :param feed_id: Returns api keys limited to that feed and its
                        datastreams.

        """
        url = self.url()
        params = {}
        if feed_id is not None:
            params['feed_id'] = feed_id
        response = self.client.get(url, params=params)
        response.raise_for_status()
        json = response.json()
        for data in json['keys']:
            key = self._coerce_key(data)
            yield key

    def get(self, key_id):
        """Fetch and return an API key by its id.

        :param key_id: The ID of the key to get.

        .. note: Unless a master API key is used, the only key that can be read
                 is this key. A master key is a non-resource restricted,
                 private key, which has permissions to perform all HTTP
                 methods.

        """
        url = self.url(key_id)
        response = self.client.get(url)
        response.raise_for_status()
        data = response.json()
        key = self._coerce_key(data['key'])
        return key

    def delete(self, key_id):
        """Delete the specified key.

        :param key_id: The key ID

        .. note: You must use a master key to delete an API Key. A master key
                 is a non-resource restricted, private key, which has
                 permissions to perform all HTTP methods.

        """
        url = self.url(key_id)
        response = self.client.delete(url)
        response.raise_for_status()

    def _coerce_key(self, data):
        api_key = data.get('api_key')
        permissions_data = data.get('permissions', [])
        data = {k: v for (k, v) in data.items() if k != 'api_key'}
        permissions = []
        for permission_data in permissions_data:
            permission = self._coerce_permission(permission_data)
            permissions.append(permission)
        data['permissions'] = permissions
        key = Key(**data)
        key._data['api_key'] = api_key
        key._manager = self
        return key

    def _coerce_permission(self, data):
        if isinstance(data, Permission):
            return data
        resources_data = data.get('resources')
        data = {k: v for (k, v) in data.items() if k != 'resources'}
        permission = Permission(**data)
        if resources_data:
            resources = []
            for resource_data in resources_data:
                resource = self._coerce_resource(resource_data)
                resources.append(resource)
            permission._data['resources'] = resources
        return permission

    def _coerce_resource(self, data):
        resource = Resource(**data)
        return resource


def _id_from_url(url):
    """Return the last part or a url

    >>> _id_from_url('http://api.xively.com/v2/feeds/1234')
    '1234'

    """
    id = url.rsplit('/', 1)[1]
    return id



# -*- coding: utf-8 -*-

import json

from datetime import datetime

try:
    from urlparse import urljoin
except ImportError:
    from urllib.parse import urljoin  # NOQA

from requests.auth import AuthBase
from requests.sessions import Session

import xively-mqtt

import paho.mqtt.client as mqtt

__all__ = ['Client']


class KeyAuth(AuthBase):
    """Attaches HTTP API Key Authentication to the given Request object."""
    def __init__(self, key):
        self.key = key

    def __call__(self, r):
        # modify and return the request
        r.headers['X-ApiKey'] = self.key
        return r


class Client(Session, mqtt):
    r"""A Xively API Client object.

    This is instantiated with an API key which is used for all requests to the
    Xively API.  It also defines a BASE_URL so that we can specify relative urls
    when using the client (all requests via this client are going to Xively).

    :param key: A Xively API Key
    :type key: str
    :param use_ssl: Use https for all connections instead of http
    :type use_ssl: bool [False]
    :param verify: Verify SSL certificates (default: True)

    A Client instance can also be used when you want low level access to the
    API and can be used with CSV or XML instead of the default JSON.

    Usage::

        >>> client = xively.Client("YOUR_API_KEY")
        >>> body = "1,123\r\n2,456\r\n"
        >>> client.post('/v2/feeds/1977.csv', data=body)
        <Response [200]>

    """
    BASE_URL = "//api.xively.com"
    MQTT_URL = "api.xively.com"
    MQTT_PORT = 1883
    mqttc = self.Client()


    def __init__(self, key, use_ssl=False, verify=True):
        super(Client, self).__init__()
        self.auth = KeyAuth(key)
        self.base_url = ('https:' if use_ssl else 'http:') + self.BASE_URL
        self.headers['Content-Type'] = 'application/json'
        self.headers['User-Agent'] = 'xively-python/{} {}'.format(
            xively.__version__, self.headers['User-Agent'])
        self._json_encoder = JSONEncoder()
        self.verify = verify

    def on_mconnect(client, userdata, rc):
        print("Connected with result code "+str(rc))

    def mconnect(self, port):
        mqttc.on_connect = on_mconnect
        mqttc.connect(MQTT_URL, MQTT_PORT, 60)
        mqttc.loop_start()

    def request(self, method, url, *args, **kwargs):
        """Constructs and sends a Request to the Xively API.

        Objects that implement __getstate__  will be serialised.

        """
        full_url = urljoin(self.base_url, url)
        if 'data' in kwargs:
            kwargs['data'] = self._encode_data(kwargs['data'])
        return super(Client, self).request(method, full_url, *args, **kwargs)

    def _encode_data(self, data, **kwargs):
        """Returns data encoded as JSON using a custom encoder.

        >>> import xively
        >>> client = Client("API_KEY")
        >>> client._encode_data({'foo': datetime(2013, 2, 22, 12, 14, 40)})
        '{"foo": "2013-02-22T12:14:40Z"}'
        >>> feed = xively.Feed(title="The Answer")
        >>> client._encode_data({'feed': feed}, sort_keys=True)
        '{"feed": {"title": "The Answer", "version": "1.0.0"}}'
        >>> datastreams = [xively.Datastream(id="1"), xively.Datastream(id="2")]
        >>> client._encode_data({'datastreams': datastreams})
        '{"datastreams": [{"id": "1"}, {"id": "2"}]}'
        """
        encoder = JSONEncoder(**kwargs) if kwargs else self._json_encoder
        return encoder.encode(data)


class JSONEncoder(json.JSONEncoder):
    """Encoder that can handle datetime objects or xively models."""

    def default(self, obj):
        if isinstance(obj, datetime):
            return obj.isoformat() + 'Z'
        elif hasattr(obj, '__getstate__'):
            return obj.__getstate__()
        else:
            return json.JSONEncoder.default(self, obj)
